import inspect
from pathlib import Path
from typing import List, Union
from RPA.Windows.keywords import (
    ActionNotPossible,
    keyword,
    LibraryContext,
)
from RPA.Windows import utils
from .locators import DEFAULT_SEARCH_TIMEOUT, WindowsElement

if utils.is_windows():
    import uiautomation as auto


class ElementKeywords(LibraryContext):
    """Keywords for handling Control elements"""

    @keyword
    def set_anchor(
        self,
        locator: Union[WindowsElement, str],
        timeout: float = DEFAULT_SEARCH_TIMEOUT,
    ):
        """Set anchor to an element specified by the locator.

        All following keywords using locators will use this element
        as an root element. Specific use case could be setting
        anchor to TableControl element and then getting column data
        belonging to that TableControl element.

        To release anchor call ``Clear Anchor`` keyword.

        :param locator: string locator or Control element
        :param timeout: timeout in seconds for element lookup (default 5.0)

        Example:

        .. code-block:: robotframework

            Set Anchor  type:Table name:Orders depth:16
            FOR  ${row}  IN RANGE  200
                ${number}=  Get Value   name:number row ${row}
                Exit For Loop If   $number == ${EMPTY}
                ${sum}=  Get Value   name:sum row ${row}
                Log   Order number:${number} has sum:{sum}
            END
            Clear Anchor
        """
        self.ctx.anchor_element = self.ctx.get_element(locator, timeout=timeout)

    @keyword
    def clear_anchor(self):
        """Clears control anchor set by ``Set Anchor``"""
        self.ctx.anchor_element = None

    @keyword
    def print_tree(
        self,
        locator: Union[WindowsElement, str] = None,
        max_depth: int = 8,
        encoding: str = "utf-8",
        capture_image_folder: str = None,
        log_as_warnings: bool = False,
    ):
        """Print Control element tree.

        Windows application structure can contain multilevel element
        structure. Understanding this structure is important for
        creating locators.

        This keyword can be used to output application element structure
        starting with the element defined by the `locator`.

        :param locator: string locator or Control element
        :param max_depth: level , defaults to 8
        :param encoding: defaults to "utf-8"
        :param capture_image_folder: if None images are not captured
        :param log_as_warnings: if set log messages are visible on the console
        """
        index = 1

        def GetFirstChild(element):
            return element.item.GetFirstChildControl()

        def GetNextSibling(element):
            return element.item.GetNextSiblingControl()

        target = self.ctx.get_element(locator)
        image_folder = (
            Path(capture_image_folder).resolve() if capture_image_folder else None
        )
        for element, depth in auto.WalkTree(
            target.item,
            getFirstChild=GetFirstChild,
            getNextSibling=GetNextSibling,
            includeTop=True,
            maxDepth=max_depth,
        ):
            element_as_text = (
                str(element).encode(encoding) if encoding else str(element)
            )
            if image_folder:
                capture_filename = f"{element.ControlType}_{index}.png"
                # TODO. exception handling
                element.CaptureToImage(str(image_folder / capture_filename))
                element_as_text += f" [{capture_filename}]"
            if log_as_warnings:
                self.ctx.logger.warning(f"{' ' * depth * 4}{element_as_text}")
            else:
                self.ctx.logger.info(f"{' ' * depth * 4}{element_as_text}")
            index += 1

    @keyword
    def get_attribute(self, locator: Union[WindowsElement, str], attribute: str):
        """Get attribute value of the element defined by the locator.

        :param locator: string locator or Control element
        :param attribute: name of the attribute to get

        Example:

        .. code-block:: robotframework

            ${id}=   Get Attribute  type:Edit name:firstname   AutomationId
        """
        # TODO. Add examples
        element = self.ctx.get_element(locator)
        attr = hasattr(element.item, attribute)
        if not attr:
            raise ActionNotPossible(
                'Element "%s" does not have "%s" attribute' % (locator, attribute)
            )
        if callable(attr):
            raise ActionNotPossible(
                'Can\'t access attribute "%s" of element "%s"' % (attribute, locator)
            )
        return str(getattr(element.item, attribute))

    @keyword
    def list_attributes(self, locator: Union[WindowsElement, str]) -> List:
        """List all element attributes.

        :param locator: string locator or Control element
        """
        element = self.ctx.get_element(locator)
        element_attributes = [e for e in dir(element.item) if not e.startswith("_")]
        attributes = []

        for attr_name in element_attributes:
            attr = getattr(element.item, attr_name)
            if not inspect.ismethod(attr):
                attributes.append(attr_name)
        return attributes



from pathlib import Path
from typing import Union, Optional

from RPA.Windows.keywords import ActionNotPossible, keyword, LibraryContext
from RPA.Windows import utils
from .locators import WindowsElement

if utils.is_windows():
    import uiautomation as auto


class ActionKeywords(LibraryContext):
    """Keywords for performing desktop actions"""

    @keyword(tags=["action", "mouse"])
    def click(self, locator: Union[WindowsElement, str]):
        """Mouse click on element matching given locator.

        :param locator: string locator or Control element

        Example:

        .. code-block:: robotframework

            Click  id:button1
            Click  id:button2 offset:10,10
            Click  name:SendButton
        """
        return self._mouse_click(locator, "Click")

    @keyword(tags=["action", "mouse"])
    def double_click(self, locator: Union[WindowsElement, str]):
        """Double mouse click on element matching given locator.

        :param locator: string locator or Control element

        Example:

        .. code-block:: robotframework

            Double Click  name:ResetButton
        """
        return self._mouse_click(locator, "DoubleClick")

    @keyword(tags=["action", "mouse"])
    def right_click(self, locator: Union[WindowsElement, str]):
        """Right mouse click on element matching given locator.

        :param locator: string locator or Control element

        Example:

        .. code-block:: robotframework

            Right Click  name:MenuButton
        """
        return self._mouse_click(locator, "RightClick")

    @keyword(tags=["action", "mouse"])
    def middle_click(self, locator: Union[WindowsElement, str]):
        """Right mouse click on element matching given locator.

        :param locator: string locator or Control element

        Example:

        .. code-block:: robotframework

            Middle Click  name:button2
        """
        return self._mouse_click(locator, "MiddleClick")

    def _mouse_click(self, element, click_type):
        element = self._get_element_for_click(element)
        if element.item.robocorp_click_offset:
            self._click_element_coordinates(element, click_type)
        else:
            self._click_element(element, click_type)
        return element

    def _click_element_coordinates(self, element, click_type):
        callable_attribute = getattr(auto, click_type, None)
        if callable_attribute:
            rect = element.item.BoundingRectangle
            offset_x, offset_y = [
                int(v) for v in element.item.robocorp_click_offset.split(",")
            ]
            callable_attribute(
                rect.xcenter() + offset_x,
                rect.ycenter() + offset_y,
                waitTime=self.ctx.wait_time,
            )
        else:
            raise ActionNotPossible(
                "Element '%s' does not have '%s' attribute" % (element, click_type)
            )

    def _click_element(self, element, click_type):
        attr = getattr(element.item, click_type, None)
        if attr and callable(attr):
            attr(waitTime=self.ctx.wait_time, simulateMove=self.ctx.simulate_move)
        else:
            raise ActionNotPossible(
                "Element '%s' does not have '%s' attribute" % (element, click_type)
            )

    def _get_element_for_click(self, locator):
        element = self.ctx.get_element(locator)
        if hasattr(element.item, "SetActive"):
            element.item.SetActive()
        element.item.MoveCursorToMyCenter(simulateMove=self.ctx.simulate_move)
        return element

    @keyword(tags=["action"])
    def select(self, locator: Union[WindowsElement, str], value: str):
        """Select value on Control element if action is supported.

        Will print warning if it is not possible to select on element.

        :param locator: string locator or Control element
        :param value: string value to select on Control element

        Example:

        .. code-block:: robotframework

            Select  type:SelectControl   option2
        """
        element = self.ctx.get_element(locator)
        if hasattr(element.item, "Select"):
            element.item.Select(value)
        else:
            raise ActionNotPossible(
                "Element '%s' does not have 'Select' attribute" % locator
            )

    @keyword(tags=["action"])
    def send_keys(
        self,
        locator: Optional[Union[WindowsElement, str]] = None,
        keys: str = None,
        interval: float = 0.01,
        wait_time: float = None,
        send_enter: bool = False,
    ):
        """Send keys to desktop, current window or to Control element
        defined by given locator.

        If ``locator`` is `None` then keys are sent to desktop.

        :param locator: string locator or Control element (default None means desktop)
        :param keys: the keys to send
        :param interval: time between sending keys, default 0.01 seconds
        :param wait_time: time to wait after sending keys, default is a
         library `wait_time`, see keyword ``Set Wait Time``
        :param send_enter: if True then {Enter} is sent at the end of the keys

        Example:

        .. code-block:: robotframework

            Send Keys  desktop   {Ctrl}{F4}
            Send Keys  keys={Ctrl}{F4}   # locator will be NONE, keys send to desktop
            Send Keys  id:input5  username   send_enter=True
            ${element}=   Get Element   id:pass
            Send Keys  ${element}  password   send_enter=True
        """
        element = self.ctx.get_element(locator)
        keys_wait_time = wait_time or self.ctx.wait_time
        if send_enter:
            keys += "{Enter}"
        if hasattr(element.item, "SendKeys"):
            element.item.SendKeys(keys, interval=interval, waitTime=keys_wait_time)
        else:
            raise ActionNotPossible(
                "Element '%s' does not have 'SendKeys' attribute" % locator
            )

    @keyword
    def get_text(self, locator: Union[WindowsElement, str]) -> str:
        """Get text from Control element defined by the locator.

        :param locator: string locator or Control element

        Example:

        .. code-block:: robotframework

            ${date}=  Get Text   type:Edit name:'Date of birth'
        """
        element = self.ctx.get_element(locator)
        if hasattr(element.item, "GetWindowText"):
            return element.item.GetWindowText()
        raise ActionNotPossible(
            "Element '%s' does not have 'GetWindowText' attribute" % locator
        )

    @keyword
    def get_value(self, locator: Union[WindowsElement, str]) -> str:
        """Get value of the element defined by the locator.

        :param locator: string locator or Control element

        Example:

        .. code-block:: robotframework

            ${value}=   Get Value   type:DataItem name:column1
        """
        element = self.ctx.get_element(locator)
        if hasattr(element.item, "GetValuePattern"):
            value_pattern = element.item.GetValuePattern()
            return value_pattern.Value
        raise ActionNotPossible(
            "Element '%s' does not have 'GetValuePattern' attribute" % locator,
        )

    @keyword(tags=["action"])
    def set_value(
        self, locator: Union[WindowsElement, str], value: str, enter: bool = False
    ) -> None:
        """Set value of the element defined by the locator.

        :param locator: string locator or Control element
        :param value: string value to set
        :param enter: set True to send ENTER key to the element

        Example:

        .. code-block:: robotframework

            Set Value   type:DataItem name:column1   ab c  # Set value to "ab c"
            # Press ENTER after setting the value
            Set Value    type:Edit name:'File name:'    console.txt    True
        """
        element = self.ctx.get_element(locator)
        if hasattr(element.item, "GetValuePattern"):
            value_pattern = element.item.GetValuePattern()
            value_pattern.SetValue(value)
            if enter:
                self.send_keys(element, "{enter}")
        else:
            raise ActionNotPossible(
                "Element '%s' does not have 'ValuePattern' attribute to set" % locator,
            )

    @keyword
    def set_wait_time(self, wait_time: float) -> float:
        """Set library wait time for action keywords.

        The wait_time is spent after each keyword performing
        mouse or keyboard action.

        Library default wait_time is `0.5`

        Returns value of the previous wait_time value.

        :param wait_time: float value (in seconds), e.g. `0.1`

        Example:

        .. code-block:: robotframework

            ${old_wait_time}=  Set Wait Time  0.2
        """
        old_value = self.ctx.wait_time
        self.ctx.wait_time = wait_time
        return old_value

    @keyword
    def screenshot(self, locator: Union[WindowsElement, str], filename: str):
        """Take a screenshot of the element defined by the locator.

        :param locator: string locator or Control element
        :param filename: image filename
        """
        filepath = Path(filename).resolve()
        element = self.ctx.get_element(locator)
        if hasattr(element.item, "CaptureToImage"):
            element.item.CaptureToImage(str(filepath))
        else:
            raise ActionNotPossible(
                "Element '%s' does not have 'CaptureToImage' attribute" % locator,
            )



from robot.api.deco import keyword

from .context import (
    ActionNotPossible,
    LibraryContext,
    ControlNotFound,
    MultipleControlsFound,
    TimeoutException,
    WindowControlError,
)
from .action import ActionKeywords
from .elements import ElementKeywords
from .locators import LocatorKeywords
from .window import WindowKeywords



from dataclasses import dataclass, field
import re
from typing import List, Union
from RPA.Windows.keywords import (
    ControlNotFound,
    keyword,
    LibraryContext,
    WindowControlError,
)
from RPA.Windows import utils

if utils.is_windows():
    import uiautomation as auto
    from uiautomation.uiautomation import Control

DEFAULT_SEARCH_TIMEOUT = 5.0

WINDOWS_LOCATOR_STRATEGIES = {
    "automationid": "AutomationId",
    "id": "AutomationId",
    "class": "ClassName",
    "control": "ControlType",
    "depth": "searchDepth",
    "name": "Name",
    "regex": "RegexName",
    "subname": "SubName",
    "type": "ControlType",
    "index": "foundIndex",
    "offset": "offset",
    "desktop": "desktop",
    "process": "process",
    "handle": "handle",
    "executable": "executable",
}


@dataclass
class WindowsElement:
    """Represent Control as dataclass"""

    item: Control
    locator: str = None
    name: str = ""
    automation_id: str = ""
    control_type: str = ""
    class_name: str = ""


@dataclass
class MatchObject:
    """Represents all locator parts as object properties"""

    match_type: str = field(default="all")
    match_index: int = None
    locators: List = field(default_factory=list)
    _classes: List = field(default_factory=list)
    regex: str = None
    regex_field: str = None
    max_level: int = 0

    def parse_locator(self, locator: str):
        locator_tree = [loc.strip() for loc in locator.split(">")]
        # self.logger.warning(locator_tree)
        regex = rf"({':|'.join(WINDOWS_LOCATOR_STRATEGIES.keys())}:|or|and)('{{1}}(.+)'{{1}})|(\S+)?"  # noqa: E501
        match_object = MatchObject()
        for level, branch in enumerate(locator_tree):
            parts = re.finditer(regex, branch, re.IGNORECASE)

            default_value = []
            strategy = None

            for part in parts:
                self.handle_locator_part(
                    level, part, match_object, default_value, strategy
                )
            if not strategy and len(default_value) > 0:
                match_object.add_locator("Name", " ".join(default_value), level)
        if len(match_object.locators) == 0:
            match_object.add_locator("Name", locator)
        return match_object

    def handle_locator_part(self, level, part, match_object, default_value, strategy):
        part_text = part.group(0)
        if len(part_text.strip()) == 0:
            return
        if part_text == "or":
            match_object.match_type = "any"
        elif part_text == "and":
            pass
        else:
            try:
                strategy, value = part_text.split(":", 1)
            except ValueError:
                strategy = None
                default_value.append(part_text)
            # self.logger.info("STRATEGY: %s VALUE: %s" % (strategy, value))
            if strategy and strategy in WINDOWS_LOCATOR_STRATEGIES:
                if len(default_value) > 0:
                    match_object.add_locator("Name", " ".join(default_value))
                    default_value = []
                windows_locator_strategy = WINDOWS_LOCATOR_STRATEGIES[strategy]
                match_object.add_locator(windows_locator_strategy, value, level)

    def add_locator(self, strategy, value, level=0) -> None:
        if not value:
            return
        self.max_level = max(self.max_level, level)
        value = value.replace("'", "").strip()
        if strategy == "regex":
            self.regex = value
        elif strategy == "regex_field":
            self.regex_field = value
        elif strategy in ["foundIndex", "searchDepth", "handle"]:
            value = int(value.strip())
            self.locators.append([strategy, value, level])
        elif strategy == "ControlType":
            value = value if value.endswith("Control") else f"{value}Control"
            self.locators.append([strategy, value, level])
        else:
            self.locators.append([strategy, value, level])
        if (
            strategy in ["class", "class_name", "friendly", "friendly_class_name"]
            and value.lower() not in self._classes
        ):
            self._classes.append(value.lower())

    @property
    def classes(self) -> List:
        uniques = []
        for c in self._classes:
            if c not in uniques:
                uniques.append(c)
        return uniques


class LocatorKeywords(LibraryContext):
    """Keywords for handling Windows locators"""

    def _get_element_with_locator_part(
        self, locator, search_depth, root_element=None
    ) -> Control:
        # TODO. refactor code
        element = None

        match_object = MatchObject()
        mo = match_object.parse_locator(locator)
        self.ctx.logger.info("locator '%s' to match element: %s" % (locator, mo))
        search_params = {}
        for loc in mo.locators:
            search_params[loc[0]] = loc[1]
        offset = search_params.pop("offset", None)
        if "searchDepth" not in search_params.keys():
            search_params["searchDepth"] = search_depth
        if "executable" in search_params.keys():
            root_element = auto.GetRootControl()
            search_params.pop("ControlType")
            executable = search_params.pop("executable")
            window_list = self.ctx.list_windows()
            matches = [w for w in window_list if w["name"] == executable]
            if not matches:
                raise WindowControlError(
                    "Could not locate window with executable '%s'" % executable
                )
            elif len(matches) > 1:
                raise WindowControlError(
                    "Found more than one window with executable '%s'" % executable
                )
            self.logger.info(
                "Found process with window title: '%s'" % matches[0]["title"]
            )
            search_params["Name"] = matches[0]["title"]
            element = Control(**search_params)
            new_element = Control.CreateControlFromControl(element)
            new_element.robocorp_click_offset = offset
            return new_element
        elif "handle" in search_params.keys():
            root_element = auto.GetRootControl()
            search_params.pop("ControlType")
            handle = search_params.pop("handle")
            window_list = self.ctx.list_windows()
            matches = [w for w in window_list if w["handle"] == handle]
            if not matches:
                raise WindowControlError(
                    "Could not locate window with handle '%s'" % handle
                )
            elif len(matches) > 1:
                raise WindowControlError(
                    "Found more than one window with handle '%s'" % handle
                )
            self.logger.info(
                "Found process with window title: '%s'" % matches[0]["title"]
            )
            search_params["Name"] = matches[0]["title"]
            element = Control(**search_params)
            new_element = Control.CreateControlFromControl(element)
            new_element.robocorp_click_offset = offset
            return new_element

        if "desktop" in search_params.keys():
            root_element = auto
            search_params.pop("desktop")
        # if "ControlType" in search_params.keys():
        #    control_type = search_params.pop("ControlType")
        #    child_control = getattr(root_control, control_type)
        #    new_control = control(**search_params)
        # new_control.robocorp_click_offset = offset
        # return new_control

        control_type = search_params.pop("ControlType", "Control")
        element = getattr(root_element, control_type)

        new_element = element(**search_params)
        new_element.robocorp_click_offset = offset
        return new_element

    @keyword
    def get_element(
        self,
        locator: Union[str, WindowsElement],
        search_depth: int = 8,
        root_element: WindowsElement = None,
        timeout: float = DEFAULT_SEARCH_TIMEOUT,
    ) -> WindowsElement:
        """Get Control element defined by the locator.

        Returned element can be used instead of a locator string for
        keywords accepting `locator`.

        Keyword ``Get Attribute`` can be used to read element attribute values.

        :param locator: locator as a string or as a element
        :param search_depth: how deep the element search will traverse (default 8)
        :param root_element: can be used to set search root element
        :param timeout: timeout in seconds for element lookup (default 5.0)

        Example:

        .. code-block:: robotframework

            Set Anchor    id:DataGrid
            ${elements}=    Get Elements    type:HeaderItem
            FOR    ${el}    IN    @{elements}
                Log To Console    ${el.item.Name}
            END
        """
        # TODO. Add examples
        self.logger.info("Locator '%s' into element", locator)
        if not locator:
            element = self.ctx.window or WindowsElement(auto.GetRootControl(), None)
        elif isinstance(locator, str):
            element = self.get_element_by_locator_string(
                locator, search_depth, root_element
            )
        else:
            element = locator
        if hasattr(element.item, "Exists"):
            element.item.Exists(maxSearchSeconds=timeout)
        if not element:
            raise ControlNotFound("Unable to get element with '%s'" % locator)
        self.logger.info("Returning element: '%s'", element)
        return element

    def get_element_by_locator_string(self, locator, search_depth, root_element):
        locators = locator.split(" > ")
        anchor = self.ctx.anchor_element.item if self.ctx.anchor_element else None
        root = root_element.item if root_element else None
        window = self.ctx.window.item if self.ctx.window else None
        root_element = root or anchor or window or auto.GetRootControl()
        element = None
        for loc in locators:
            self.logger.info("Root element: '%s'" % root_element)
            element = self._get_element_with_locator_part(
                loc, search_depth, root_element
            )
            root_element = element
        return WindowsElement(
            element,
            locator,
            element.Name,
            element.AutomationId,
            element.ControlTypeName,
            element.ClassName,
        )

    @keyword
    def get_elements(
        self,
        locator: Union[str, WindowsElement],
        search_depth: int = 8,
        root_element: WindowsElement = None,
        timeout: float = DEFAULT_SEARCH_TIMEOUT,
    ) -> List:
        """Get list of elements matching locator.

        :param locator: locator as a string or as a element
        :param search_depth: how deep the element search will traverse (default 8)
        :param root_element: can be used to set search root element
        :param timeout: timeout in seconds for element lookup (default 5.0)

        Example:

        .. code-block:: robotframework

            Set Anchor    id:DataGrid
            ${elements}=    Get Elements    type:HeaderItem
            FOR    ${el}    IN    @{elements}
                Log To Console    ${el.Name}
            END
        """
        elements = []
        window_element = self.get_element(locator, search_depth, root_element, timeout)
        elements.append(window_element)
        while True:
            # TODO. the sibling needs to match the original locator
            next_element = window_element.item.GetNextSiblingControl()
            if next_element:
                window_element = WindowsElement(
                    next_element,
                    locator,
                    next_element.Name,
                    next_element.AutomationId,
                    next_element.ControlTypeName,
                    next_element.ClassName,
                )
                elements.append(window_element)
            else:
                break
        return elements



from datetime import datetime

# pylint: disable=C0415
from pynput_robocorp import mouse, keyboard

from RPA.Windows import utils


if utils.is_windows():
    import uiautomation as auto


record = False
recording_time = None
recording = []


def inspect_element(action: str = "Click", control_window: bool = True):
    """Inspect Windows element under mouse pointer

    :param action: which action is attached to locator
    :param control_window: set False to not include ``Control Window`` keyword
    """
    # TODO. Add Python syntax
    # TODO. Add examples
    output = []
    with auto.UIAutomationInitializerInThread(debug=False):
        control = auto.ControlFromCursor()
        parent_control = control.GetParentControl()
        top_level_control = control.GetTopLevelControl()

        parent_locator = _get_element_key_properties(parent_control)
        child_locator = _get_element_key_properties(control)
        locator_path = f"{parent_locator} > {child_locator}"
        if "name:" in child_locator:
            locator_path = child_locator
        elif "id:" in child_locator:
            locator_path = child_locator
        if control_window:
            output.append(f"Control Window  {top_level_control.Name}")
        if action:
            output.append(f"{action}  {locator_path}")
        else:
            output.append(locator_path)
        if record:
            # skip similar locators
            # unique = True
            # for item in self.recording:
            #     if (
            #         item["top"] == top_level_control.Name
            #         and item["locator"] == locator_path
            #     ):
            #         unique = False
            #         break
            # if unique:
            recording.append(
                {
                    "type": "locator",
                    "top": top_level_control.Name,
                    "top_handle": top_level_control.NativeWindowHandle,
                    "x": top_level_control,
                    "locator": locator_path,
                }
            )
    return output


def _get_element_key_properties(element, regex_limit=300):
    automation_id = element.AutomationId
    element_name = element.Name
    snippet = ""
    if element_name and len(element_name) > 0:
        snippet = f"name:'{element_name.strip()}'"
    elif automation_id and not utils.is_integer(automation_id):
        snippet = f"id:{automation_id}"
    else:
        control_type = element.ControlTypeName.strip()
        class_name = element.ClassName.strip()
        name = element.Name.strip()
        name_property = "name:"
        if len(name) > regex_limit:
            name_property = "regex:"
            name = name[:regex_limit].strip()
        locators = []
        if len(control_type) > 0:
            locators.append(f"type:{control_type}")
        if len(class_name) > 0:
            locators.append(f"class:{class_name}")
        if len(name) > 0 and " " in name:
            locators.append(f"{name_property}'{name}'")
        elif len(name) > 0:
            locators.append(f"{name_property}{name}")
        if not locators:
            print("Was unable to construct locator for the control")
            return None
        else:
            snippet = " and ".join(locators)
    return snippet


def stop_listeners():
    mouse.Listener.stop()
    keyboard.Listener.stop()


def start_recording():
    """Start recording mouse clicks.

    Can be stopped by pressing keyboard ``ESC``.
    """
    # TODO. Add examples
    global record, recording  # pylint: disable=W0603
    record = True
    recording = []

    def on_click(x, y, button, pressed):  # pylint: disable=W0613
        global recording, recording_time  # pylint: disable=W0602
        if pressed:
            inspect_time = datetime.now()
            if recording_time:
                timediff = inspect_time - recording_time
                seconds = max(round(float(timediff.microseconds / 1000000.0), 1), 0.1)
                recording.append({"type": "sleep", "value": seconds})
            recording_time = inspect_time
            inspect_element()

    def on_release(key):
        if key == keyboard.Key.esc:
            return False
        return True

    mouse_listener = mouse.Listener(on_click=on_click)
    mouse_listener.start()
    with keyboard.Listener(on_release=on_release) as key_listener:
        print("keyboard and mouse listeners started", flush=True)
        key_listener.join()
    mouse_listener.stop()
    the_recording = get_recording()
    print(the_recording)


def get_recording(sleeps=False):
    """Get list of recorded steps.

    :param sleeps: set False to exclude recording sleeps
    """
    # TODO. atm will always use CLICK
    # TODO. Add examples
    global recording  # pylint: disable=W0602
    output = []
    top = None
    action_name = "Click"
    for item in recording:
        if sleeps and item["type"] == "sleep":
            output.append(f"Sleep   {item['value']}s")
        if (
            item["type"] == "locator"
            and not top
            or "top" in item.keys()
            and item["top"] != top
        ):
            output.append(
                f"Control Window    {item['top']}  # Handle: {item['top_handle']}"
            )
            top = item["top"]
        if item["type"] == "locator":
            output.append(f"{action_name}   {item['locator']}")

    result = "\n".join(output)
    header = (
        f"\n{'-'*80}"
        "\nCOPY & PASTE BELOW CODE INTO *** Tasks *** or *** Keywords ***"
        f"\n{'-'*80}\n\n"
    )
    footer = f"\n\n{'-'*80}"
    return f"{header}{result}{footer}"


def main():
    start_recording()


if __name__ == "__main__":
    main()



import os
import signal
import time
from typing import List, Dict

from RPA.Windows.keywords import keyword, LibraryContext, WindowControlError
from RPA.Windows import utils
from .locators import WindowsElement

if utils.is_windows():
    import uiautomation as auto


class WindowKeywords(LibraryContext):
    """Keywords for handling Window controls"""

    @keyword(tags=["window"])
    def control_window(self, locator: str = None) -> int:
        """Controls the window defined by the locator.

        This means that this window is used as a root element
        for all the following keywords using locators.

        Returns native handle of the window.

        :param locator: string locator

        Example:

        .. code-block:: robotframework

            Control Window   Calculator
            Control Window   name:Calculator
            Control Window   subname:Notepad
            Control Window   regex:.*Notepad
            ${handle}=  Control Window   executable:Spotify.exe
        """
        window_locator = f"{locator}  and type:WindowControl"
        pane_locator = f"{locator}  and type:PaneControl"
        desktop = WindowsElement(auto.GetRootControl(), window_locator)
        self.ctx.window = self.ctx.get_element(window_locator, root_element=desktop)
        if not self.ctx.window.item.Exists():
            desktop = WindowsElement(auto.GetRootControl(), pane_locator)
            self.ctx.window = self.ctx.get_element(pane_locator, root_element=desktop)
        if not self.ctx.window.item.Exists():
            raise WindowControlError(
                'Could not locate window with locator "%s"' % locator
            )
        # auto.WaitForExist(self.ctx.window, 5)
        if hasattr(self.ctx.window.item, "Restore"):
            self.ctx.window.item.Restore()
        self.ctx.window.item.SetFocus()
        self.ctx.window.item.SetActive()
        handle = self.ctx.window.item.NativeWindowHandle
        self.ctx.window.item.MoveCursorToMyCenter(simulateMove=self.ctx.simulate_move)
        # time.sleep(1.0)
        return handle

    @keyword(tags=["window"])
    def minimize_window(self, locator: str = None) -> None:
        """Minimize the current active window or the window defined
        by the locator.

        :param locator: string locator

        Example:

        .. code-block:: robotframework

            Minimize Window   # Current active window
            Minimize Window   executable:Spotify.exe
        """
        if locator:
            self.control_window(locator)
        if not self.ctx.window:
            raise WindowControlError("There is no active window")
        if not hasattr(self.ctx.window.item, "Minimize"):
            self.logger.warning(
                "Control '%s' does not have attribute Minimize" % self.ctx.window
            )
            return
        self.ctx.window.item.Minimize()

    @keyword(tags=["window"])
    def maximize_window(self, locator: str = None) -> None:
        """Minimize the current active window or the window defined
        by the locator.

        :param locator: string locator

        Example:

        .. code-block:: robotframework

            Maximize Window   # Current active window
            Maximize Window   executable:Spotify.exe
        """
        if locator:
            self.control_window(locator)
        if not self.ctx.window:
            raise WindowControlError("There is no active window")
        if not hasattr(self.ctx.window.item, "Maximize"):
            raise WindowControlError("Window does not have attribute Maximize")
        self.ctx.window.item.Maximize()

    @keyword(tags=["window"])
    def list_windows(self) -> List[Dict]:
        """List all window element on the system.

        :return: list of dictionaries containing information
         about Window elements

        Example:

        .. code-block:: robotframework

            ${windows}=  List Windows
            FOR  ${window}  IN  @{windows}
                Log  Window title:${window}[title]
                Log  Window process name:${window}[name]
                Log  Window process id:${window}[pid]
                Log  Window process handle:${window}[handle]
            END
        """
        windows = auto.GetRootControl().GetChildren()
        process_list = utils.get_process_list()
        win_list = []
        for win in windows:
            pid = win.ProcessId
            info = {
                "title": win.Name,
                "pid": win.ProcessId,
                "name": process_list[pid] if pid in process_list.keys() else None,
                "handle": win.NativeWindowHandle,
            }
            win_list.append(info)
        return win_list

    @keyword(tags=["window"])
    def windows_run(self, text: str, wait_time: float = 3.0) -> None:
        """Use Windows run window to launch application.

        Activated by pressing `win + r`.

        :param text: text to enter into run input field
        :param wait_time: sleep time after search has been entered (default 3.0 seconds)

        Example:

        .. code-block:: robotframework

            Windows Run   explorer.exe
        """
        self.ctx.send_keys(None, "{Win}r")
        self.ctx.send_keys(None, text)
        self.ctx.send_keys(None, "{Enter}")
        time.sleep(wait_time)

    @keyword(tags=["window"])
    def windows_search(self, text: str, wait_time: float = 3.0) -> None:
        """Use Windows search window to launch application.

        Activated by pressing `win + s`.

        :param text: text to enter into search input field
        :param wait_time: sleep time after search has been entered (default 3.0 seconds)

        Example:

        .. code-block:: robotframework

            Windows Search   Outlook
        """
        self.ctx.send_keys(None, "{Win}s")
        self.ctx.send_keys(None, text)
        self.ctx.send_keys(None, "{Enter}")
        time.sleep(wait_time)

    @keyword(tags=["window"])
    def close_current_window(self) -> bool:
        """Closes current active window or logs a warning message.

        Example:

        .. code-block:: robotframework

            Close Current Window
        """
        if not self.ctx.window:
            self.ctx.logger.warning("There is no active window")
            return False
        pid = self.ctx.window.item.ProcessId
        name = self.ctx.window.item.Name
        self.ctx.logger.info(
            'Closing window with Name:"%s", ProcessId: %s' % (name, pid)
        )
        os.kill(pid, signal.SIGTERM)
        return True
